/*
   This program demonstrates "points at infinity" in perspective projection.

   This program draws a grid of horizontal rectangles that extend deep
   into space (towards infinity). The rectangles appear to all be
   converging on the same point at infinity. You can point the rectanges
   at any "point at infinity".
*/
#include <windows.h>
#include "glut.h"
#include "glui.h"

/* GLUT callback function prototypes */
void display(void);
void reshape(int, int);
void slider1(double);
void slider2(double);
void slider3(double);

int windowHeight;         // this is set by the glut reshape() callback function
int windowWidth;          // this is set by the glut reshape() callback function

#define sliderInit 0.5
double slider1Value = sliderInit;  // this is set by the glui slider() callback function
double slider2Value = sliderInit;
double slider3Value = sliderInit;

int main(int argc, char** argv)
{
   int window;
   /* GLUT initialization */
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);  // glui needs double buffering!
   glutInitWindowSize(500,520);       // width, height (in pixels)
   glutInitWindowPosition(100,150);   // left, top (in pixels)
   window = glutCreateWindow(argv[0]);
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);

   // parent window, x, y, width, height, initial value, callback function
   gluiHorizontalSlider(window, 10,  0, -10, 20, sliderInit, slider1);
   gluiHorizontalSlider(window, 10, 20, -10, 20, sliderInit, slider2);
   gluiHorizontalSlider(window, 10, 40, -10, 20, sliderInit, slider3);

   /* Initialize OpenGL state */
   glClearColor(1.0, 1.0, 1.0, 0.0); /* white background */
   glLineWidth(1.0);

   glutMainLoop(); /* Enter event loop */
   return 0;       /* We never get here */
}//main()

void model( void )
{
   glPushMatrix();
   glRotatef((0.5-slider2Value)*180.0, 0.0, 1.0, 0.0); // rotate side to side
   glRotatef((0.5-slider3Value)*180.0, 1.0, 0.0, 0.0); // rotate up and down
   glScalef(1.0, 1.0, slider1Value*2000.0); // scale the cube "backwards" only
   glTranslatef(0.0, 0.0, -0.5);  // push the glut cube backwards so that its face is in the xy-plane
   glutWireCube(1.0);
   //glutSolidCube(1.0);
   glPopMatrix();
}
//model()


void display(void)
{
   int i, j;

   // specify the viewport
   glViewport( 0,  0, windowWidth, windowHeight-20); // left, bottom, width, height (in pixels)

   /* set up viewing */
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glFrustum(-5.0, 5.0, -5.0, 5.0, 9.0, 11.0+slider1Value*2000.0);

   glClear( GL_COLOR_BUFFER_BIT );

   glColor3f(0.0f, 1.0f, 0.0f);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
//   glTranslatef(0.0, 0.0, -1.0);  // step back away from the camera
//   glTranslatef(0.0, 0.0, -1.1);  // step back away from the camera (get rid of rounding errors???)
glTranslatef(0.0, 0.0, -10.0);
   glTranslatef(-4.0, 4.0, 0.0);  // put ourselves in at the upper left hand corner of the picture
   for (i = 0; i < 5; i++)
   {
      glPushMatrix();
      for (j = 0; j < 5; j++)
      { if ((j==0 && i==0) || (j==4 && i==4) || (j==0 && i==4) || (j==4 && i==0) ||  (j==2 && i==2))
         {
            model();
          }
            glTranslatef(2.0, 0.0, 0.0);  // move over one column (these translations accumulate)

      }
      glPopMatrix();                   // go back to the left hand column
      glTranslatef(0.0, -2.0, 0.0);    // drop down one row (these translations accumulate)
   }

   glutSwapBuffers();  // glui needs double buffering!
}//display()


/* The reshape() callback function */
void reshape(int width, int height)
{
   windowWidth  = width;
   windowHeight = height;

   printf("We got a \"reshape(%d, %d)\" event\n", width, height);

   gluiReshape(width, height);  // notice the difference here
}//reshape()


/* glui slider callback function */
void slider1(double value)
{
   slider1Value = value;

   printf("We got a \"slider1(%f)\" event\n", value);

   glutPostRedisplay();  // notice the difference here
}//slider()


void slider2(double value)
{
   slider2Value = value;

   printf("We got a \"slider2(%f)\" event\n", value);

   glutPostRedisplay();  // notice the difference here
}//slider2()

void slider3(double value)
{
   slider3Value = value;

   printf("We got a \"slider3(%f)\" event\n", value);

   glutPostRedisplay();  // notice the difference here
}//slider3()
